/***********************************************************************************************

      File: 
     Title: APT Automatic Parallelizing Tool
    Author: Nils Schmeisser
   Version: 1.0
  Rev.Date: 20.2.1996

   Comment: 

***********************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <dep.h>
#include <profiler.h>
#include <cppmacro.h>
#include <apt.h>
#include <aptparser.h>

static FILE *depinfo;
static ciliststruct *deproot=NULL;
static ciliststruct *deplast=NULL;
static unsigned long int ndep=0;
FILE *depfile;

void InitDepChecker() {

  depinfo=fopen("aptdep.dat","r");
  if (depinfo==NULL) {
    APTDepCheckerError("Can't find aptdep.dat Dependency Checker Information File");
    exit(-1);
  }
  depfile=fopen("apt.depfile","w");
  fprintf(depfile,"APT Automatic Parallelization Tool, Runtime System V 1.0\n");
  fprintf(depfile,"(C) Nils Schmeisser, FVTK FZ-Rossendorf, 1996\n");
  fprintf(depfile,"\n");
  fprintf(depfile,"Dependency Protocol for 'apt.exe'\n");
  fprintf(depfile,"Date: \n");
  fprintf(depfile,"\n");
  fprintf(depfile,"--------------------------------------------------------\n");
  fprintf(depfile,"\n");
}

codeinfostruct **FindCodeInfo(int what,codeinfostruct mask,int *n) {
  ciliststruct *p;
  codeinfostruct **r;
  int i,j,found,f;

  *n=0;
  r=new(codeinfostruct*,ndep);
  p=deproot;
  while (p!=NULL) {
    found=0;
    if (mask.type>0) {
      f=(mask.type==p->item->type);
      found=(what & SAND) ? (found & f):found;
      found=(what & SOR) ? (found | f):found;
    }
    if (strlen(mask.name)>0) {
      f=(strcmp(mask.name,p->item->name)!=0);
      found=(what & SAND) ? (found & f):found;
      found=(what & SOR) ? (found | f):found;
    }
    if (strlen(mask.modname)>0) {
      f=(strcmp(mask.modname,p->item->modname)!=0);
      found=(what & SAND) ? (found & f):found;
      found=(what & SOR) ? (found | f):found;
    }
    if (mask.line1>0) {
      f=((mask.line1>=p->item->line1) && (mask.line1<=p->item->line2));
      found=(what & SAND) ? (found & f):found;
      found=(what & SOR) ? (found | f):found;
    }
    if (mask.line2>0) {
      f=((mask.line2>=p->item->line1) && (mask.line2<=p->item->line2));
      found=(what & SAND) ? (found & f):found;
      found=(what & SOR) ? (found | f):found;
    }
    if ((mask.ninput>0) && (p->item->ninput>0)) {
      for (i=0;i<mask.ninput;i++) {
	for (j=0;j<p->item->ninput;j++) {
	  f=(strcmp(mask.input[i],p->item->input[j])!=0);
	  found=(what & SAND) ? (found & f):found;
	  found=(what & SOR) ? (found | f):found;
	}
      }
    }
    if ((mask.noutput>0) && (p->item->noutput>0)) {
      for (i=0;i<mask.noutput;i++) {
	for (j=0;j<p->item->noutput;j++) {
	  f=(strcmp(mask.output[i],p->item->output[j])!=0);
	  found=(what & SAND) ? (found & f):found;
	  found=(what & SOR) ? (found | f):found;
	}
      }
    }
    if (found) {
      r[*n]=p->item;
      (*n)++;
    }
  }
  if (n!=0)
    return r;
  else {
    delete(r);
    return NULL;
  }
}

void ParseInputOutput(char *code,codeinfostruct  *ci) {
  int i,j,k,l,nl,cl;
  int inset;
  char **line;
  char **input;
  char **output;
  int ninput,noutput;
  char *s;
  char msg[1024];

  sprintf(msg,"analysing line %d .. %d",ci->line1,ci->line2);
  APTMessage(msg);
  nl=0;
  cl=strlen(code);
  for (i=0;i<cl;i++) if (code[i]==0x0a) nl++;
  if (code[cl]!=0x0a) nl++;
  line=new(char*,nl);
  s=code;
  j=0;
  for (i=0;i<nl;i++) {
    line[i]=new(char,1024);
    line[i][0]=0;
    k=0;
    while ((j<cl) && (code[j]!=0x0a) && (k<1024)) {
      line[i][k]=code[j];
      k++;
      j++;
    }
    j++;
    if (k>1024) {
      s=new(char,256);
      sprintf(s,"line %d too large, cut to less then 1024 characters",ci->line1+i);
      APTError(s);
      exit(-1);
    }
    line[i][k]=0;
  }

  //--- got all lines, now parsing each line

  input =new(char*,1024);
  output=new(char*,1024);
  for (i=0;i<1024;i++) {
    input[i]=NULL;
    output[i]=NULL;
  }
  for (i=0;i<nl;i++) if (strlen(line[i])>0) {
    ninput=0;
    noutput=0;
    for (k=0;k<1024;k++) {
      if (input[i]!=NULL) {
	delete(input[i]);
	input[i]=NULL;
      }
      if (output[i]!=NULL) {
	delete(output[i]);
	output[i]=NULL;
      }
    }
    if (ParseLine(SOURCE,line[i],input,&ninput,output,&noutput)) {

      //--- these lines are for control reasons during debugging

      fprintf(depfile,"%s:%d:INPUT: ",ci->modname,ci->line1+i);
      for (k=0;k<ninput;k++)
	fprintf(depfile," %s",input[k]);
      fprintf(depfile,"\n");
      fprintf(depfile,"%s:%d:OUTPUT: ",ci->modname,ci->line1+i);
      for (k=0;k<noutput;k++)
	fprintf(depfile," %s",output[k]);
      fprintf(depfile,"\n");

      //--- for all input search for occurency in ci->input, if not insert

      for (k=0;k<ninput;k++) {
	inset=0;
	for (l=0;l<ci->ninput;l++)
	  if (!(strcmp(input[k],ci->input[l])))
	    inset=1;
	if (!inset) {
	  ci->input[ci->ninput]=new(char,strlen(input[k])+1);
	  strcpy(ci->input[ci->ninput],input[k]);
	  ci->ninput++;
	}
      }
 
      //--- for all output search for occurency in ci->output, if not insert

      for (k=0;k<noutput;k++) {
	inset=0;
	for (l=0;l<ci->noutput;l++)
	  if (!(strcmp(output[k],ci->output[l])))
	    inset=1;
	if (!inset) {
	  ci->output[ci->noutput]=new(char,strlen(output[k])+1);
	  strcpy(ci->output[ci->noutput],output[k]);
	  ci->noutput++;
	}
      }
 
    } else {
      s=new(char,256);
      sprintf(s,"Syntax error in line %d, anyway can't analyze statement",ci->line1+i);
      APTError(s);
      exit(-1);
    }
  }
}

codeinfostruct *GenerateCodeInfo(char *code,
				 char *modname,unsigned long line1,unsigned long line2,
				 unsigned long *count,int type,int profiling,FILE *f) {
  codeinfostruct *ci;
  char *s;
  char *t,*h;
  char **cline;
  unsigned long linecount,i;
  int generateinfo=1;
  int result;

  if (strlen(code)==0) generateinfo=0;
  if (profiling & NO_CODE_INFO) generateinfo=0;
  s=new(char,1024);
  if (generateinfo) {
//    sprintf(s,"building dependencies in (%d,%d) in %s",line1,line2,modname);
//    APTMessage(s);
  }
  rewind(depinfo);
  do {
    result=GetLine(depinfo,s);
  } while ((strcmp(s,"Generate_No_Code_Info")) && result!=0);
  if (result==0) {
    APTDepCheckerError("Corrupted Information file");
    exit(-1);
  }
  if (PMODE==ALL_MODE) {
    do {
      result=GetLine(depinfo,s);
      t=NULL;
      switch (SOURCE) {
        case C_LANGUAGE:       if (strstr(s,"C:")!=NULL) t=s; break;
	case FORTRAN_LANGUAGE: if (strstr(s,"F:")!=NULL) t=s; break;
	case PASCAL_LANGUAGE:  if (strstr(s,"P:")!=NULL) t=s; break;
	case ASM_LANGUAGE:     if (strstr(s,"A:")!=NULL) t=s; break;
      }
      if (t!=NULL) {
	while (t[0]!=':') t++;
	t++;
	if (strstr(code,t)!=NULL) generateinfo=0;
      }
    } while ((result!=0) && (strstr(s,"End_Section")==NULL) && (generateinfo==1));
  }

  if (generateinfo) {
    if (deproot==NULL) {
      deproot=new(ciliststruct,1);
      CODEINFO=deproot;
      deplast=deproot;
    } else {
      deplast->next=new(ciliststruct,1);
      deplast=deplast->next;
    }
    deplast->next=NULL;
    ndep++;
    ci=new(codeinfostruct,1);
    ci->type=type;
    switch (type) {
      case CODE_STATEMENT: sprintf(s,"C_STAT_%d",count[0]++); break;
      case CODE_SEQUENCE:  sprintf(s,"C_SEQ_%d",count[1]++); break;
      case CODE_LOOP:      sprintf(s,"C_LOOP_%d",count[2]++); break;
      case CODE_PROCEDURE: sprintf(s,"C_PROC_%d",count[3]++); break;
      case CODE_FUNCTION:  sprintf(s,"C_FCT_%d",count[4]++); break;
    }
    ci->name=new(char,strlen(s)+1);
    strcpy(ci->name,s);
    ci->modname=new(char,strlen(modname)+1);
    strcpy(ci->modname,modname);
    ci->code=new(char,strlen(code)+1);
    strcpy(ci->code,code);
    ci->node=-1;
    ci->root=0;
    ci->nsucc=0;
    ci->succ=new(void*,MAXDEPNODES);
    ci->npred=0;
    ci->pred=new(void*,MAXDEPNODES);
    ci->nsyncs=0;
    ci->syncs=new(int,MAXDEPNODES);
    ci->line1=line1;
    ci->line2=line2;
    ci->time=0;
    ci->ninput=0;
    ci->input=new(char*,MAXDEPVARS);
    ci->noutput=0;
    ci->output=new(char*,MAXDEPVARS);
    linecount=line2-line1+1;
    //--- jetzt input,output bestimmen -------------
    ParseInputOutput(code,ci);
    //----------------------------------------------
    deplast->item=ci;
  }

  if (strlen(code)==0)
    fprintf(f,"\n");
  else {
    if (profiling & PROFILING) {
      if (generateinfo) {
	gettime1(f);
	fprintf(f,"%s\n",code);
	gettime2(f);
	difftime(f,ci->modname,ci->name);
      } else {
	t=NULL;
	if (!(strcmp(modname,MAINMODULE))) {
	  switch (SOURCE) {
	    case C_LANGUAGE: t=strstr(code,"main(");
	                     if (t!=NULL) {
			       while ((t[0]!='{') && (strlen(t)>0)) t++;
			       t++;  // point 1 char after opening brace
			       s=new(char,strlen(code)+strlen("\n  APTProfilerInit();\n")+1);
			       if (t[0]!=0) {
				 t[0]=0;
				 t++;
			       }
			       sprintf(s,"%s\n  APTProfilerInit();\n%s",code,t);
			     }
	                     break;
            case FORTRAN_LANGUAGE: t=strstr(code,"C_APT_MAIN:");
                                   if (t!=NULL) {
                                     while ((t[0]!=':') && (strlen(t)>0)) t++;
                                     t++;  // point 1 char after colon
                                     s=new(char,strlen(code)+strlen("\n      REAL APT_PROF_TIME(2)\n      CALL APTPROFILERINIT\n")+1);
				     if (t[0]!=0) {
				       t[0]=0;
				       t++;
				     }
                                     sprintf(s,"%s\n      REAL APT_PROF_TIME(2)\n      CALL APTPROFILERINIT\n%s",code,t);
                                   }
                                   break;
	  }
	  if (t==NULL)
	    fprintf(f,"%s\n",code);
	  else
	    fprintf(f,"%s\n",s);
	} else
	  fprintf(f,"%s\n",code);
      }
    } 
  }
  return ci;
}

int readline(FILE *f,char *s) {
  char c;
  int i,result;
  
  i=0;
  s[i]=0;
  do {
    fscanf(f,"%c",&c);
    if ((c!=0x0a) && (result!=EOF)) {
      s[i]=c;
      i++;
      s[i]=0;
    }
  } while ((c!=0x0a) && (result!=EOF));
  return result;
}

int CheckFlowDependency(codeinfostruct *s1,codeinfostruct *s2) {
  int i1,i2;

  for (i1=0;i1<s1->noutput;i1++) {
    for (i2=0;i2<s2->ninput;i2++) {
      if (!(strcmp(s1->output[i1],s2->input[i2])))
	return 1; // there is a dependency
    }
  }
  return 0;
}

int CheckAntiDependency(codeinfostruct *s1,codeinfostruct *s2) {
  int i1,i2;

  for (i1=0;i1<s1->ninput;i1++) {
    for (i2=0;i2<s2->noutput;i2++) {
      if (!(strcmp(s1->input[i1],s2->output[i2])))
	return 1; // there is a dependency
    }
  }
  return 0;
}

int CheckOutputDependency(codeinfostruct *s1,codeinfostruct *s2) {
  int i1,i2;

  for (i1=0;i1<s1->noutput;i1++) {
    for (i2=0;i2<s2->noutput;i2++) {
      if (!(strcmp(s1->output[i1],s2->output[i2])))
	return 1; // there is a dependency
    }
  }
  return 0;
}

int DataDependend(codeinfostruct *s1,codeinfostruct *s2) {
  int flow,anti,output;

  flow=CheckFlowDependency(s1,s2);
  anti=CheckAntiDependency(s1,s2);
  output=CheckOutputDependency(s1,s2);
//printf("DataDependend(%s,%s)=%d\n",s1->name,s2->name,flow || anti || output);
  return (flow || anti || output);
}

int CheckSync(codeinfostruct *s1,codeinfostruct *s2) {
  int i1,i2;

  for (i1=0;i1<s1->nsucc;i1++) {
    for (i2=0;i2<s2->nsucc;i2++) {
      if (s1->succ[i1]==s2->succ[i2])
	return 1; // there is a common successor
    }
  }
  return 0;
}

/* check, if s2 is a predcessor of s1 */

int Pred(codeinfostruct *s1,codeinfostruct *s2) {
  int i,pred;

  for (i=0;i<s2->nsucc;i++) {
    if (s1==s2->succ[i])
      return 1;
    else {
      pred=Pred(s1,(codeinfostruct*)(s2->succ[i]));
      if (pred)
	return 1;
    }
  }
  return 0;
}

void BuildDependencies(ciliststruct *cl) {
  ciliststruct *l,*n;
  int i,j,result;
  char s[1024],m[1024];
  FILE *f;

  APTMessage("reading profiler information");

  f=fopen("apt.profile","r");
  l=cl;
  while (l!=NULL) {
    rewind(f);
    do {
      result=readline(f,s);
    } while ((strncmp(s,l->item->name,strlen(l->item->name))) && (result!=EOF));
    sprintf(m,"%s@%s: %%d",l->item->name,l->item->modname);
    sscanf(s,m,&(l->item->time));

    printf("{");
    for (i=0;i<l->item->ninput;i++)
      printf("%s ",l->item->input[i]);
    printf("} -> %s -> {",l->item->name);
    for (i=0;i<l->item->noutput;i++)
      printf("%s ",l->item->output[i]);
    printf("}\n");

    l=l->next;
  }
  fclose(f);

  APTMessage("building dependency graph");

  l=cl;
  while (l!=NULL) {
    n=l->next;
    while (n!=NULL) {
      if (!DataDependend(l->item,n->item)) {
	l->item->succ[l->item->nsucc]=n->item;
	l->item->nsucc++;
	n=n->next;
      } else {
	if (n->item->npred==MAXDEPNODES) {
	  sprintf(s,"too many dependencies (maximum resolvable=%d)",MAXDEPNODES);
	  APTError(s);
	  exit(-1);
	}
	n->item->pred[n->item->npred]=l->item;
	n->item->npred++;
	l->item->succ[l->item->nsucc]=n->item;
	l->item->nsucc++;
	n=n->next;
      }
    }
    l=l->next;
  }

}

