/***********************************************************************************************

      File: 
     Title: APT Automatic Parallelizing Tool
    Author: Nils Schmeisser
   Version: 1.0
  Rev.Date: 20.2.1996

   Comment: 

***********************************************************************************************/

#include <stdio.h>
#include <cppmacro.h>

typedef struct {
          char name[20];
	  int nindep;
	  char **indep;
	  char noutdep;
	  char **outdep;
        } depknot;

typedef struct {
          depknot *item;
	  void *next;
	} liststruct;

liststruct *lptr(liststruct *l,int index) {
  liststruct *p;
  int i;

  p=l;
  for (i=0;i<index;i++) p=p->next;
  return p;
}

int **CreateInzMatrix(liststruct *knot,int n) {
  int i,j;
  int **imatrix;
  int k,l;

  imatrix=new(int*,n);
  for (i=0;i<n;i++) {
    imatrix[i]=new(int,n);
    for (j=0;j<n;j++) {
      imatrix[i][j]=0;
      // check for matching input dependencies in knot j
      for (k=0;k<lptr(knot,i)->item->nindep;k++) {
	for (l=0;l<lptr(knot,j)->item->noutdep;l++) {
	  if (!(strcmp(lptr(knot,i)->item->indep[k],lptr(knot,j)->item->outdep[l]))) imatrix[i][j]=1;
	}
      }
    }
  }
  return imatrix;
}

depknot *ScanDepKnot(char *s) {
  depknot *knot;
  int i;
  char *t;

  knot=new(depknot,1);
  knot->nindep=0;
  knot->noutdep=0;
  i=0;
  while ((i<strlen(s)) && (s[i]!=':')) i++;
  strncpy(knot->name,s,i);
printf(" knotname: %s\n",knot->name);
  s+=i+1;
  t=new(char,strlen(s)+1);
  strcpy(t,s);

  while (s[0]!='-') {
    while (s[0]==' ') s++;
    if (s[0]!='-') {
      knot->nindep++;
      while (s[0]!=' ') s++;  // this is the same as reading sscanf(s,"%s",knot.inpdep[i]);
    }
  }

  s+=2;      // point to space after ->

  while (strlen(s)>0) {
    while (s[0]==' ') s++;
    knot->noutdep++;
    while ((strlen(s)>0) && (s[0]!=' ')) s++; // same as reading sscanf(s,"%s",knot.outdep[i]);
  }

  // now allocating space and reading input/output dependencies

  knot->indep =new(char*,knot->nindep);
  for (i=0;i<knot->nindep;i++) knot->indep[i]=new(char,32);
  knot->outdep=new(char*,knot->noutdep);
  for (i=0;i<knot->noutdep;i++) knot->outdep[i]=new(char,32);

  i=0;
printf(" inputdep: ");
  while (t[0]!='-') {
    while (t[0]==' ') t++;
    if (t[0]!='-') {
      sscanf(t,"%s",knot->indep[i]);
printf("%s ",knot->indep[i]);
      i++;
      while (t[0]!=' ') t++;
    }
  }
printf("\n");

  t+=2;      // point to space after ->

printf("outputdep: ");
  i=0;
  while (strlen(t)>0) {
    while (t[0]==' ') t++;
    sscanf(t,"%s",knot->outdep[i]);
    while ((strlen(t)>0) && (t[0]!=' ')) t++;
printf("%s ",knot->outdep[i]);
    i++;
  }
printf("\n");
  return knot;

}

int ScanFLine(FILE *f,char *s) {
  int i;
  char c;
  int result;

  i=0;
  do {
    result=fread(&c,1,1,f);
    if (c!=0x0a) {
      s[i]=c;
      i++;
      s[i]=0;
    }
  } while ((result!=0) && (c!=0x0a));
  return result;
}

liststruct *ScanFile(char *fname,int *n) {
  liststruct *list,*p;
  FILE *f;
  char s[1024];
  int result;

  *n=0;
  list=NULL;
  p=NULL;
  if ((f=fopen(fname,"r"))!=NULL) {
    do {
      result=ScanFLine(f,s);
      if (result!=0) {
	if (p==NULL) {
	  list=new(liststruct,1);
	  p=list;
	} else {
	  p->next=new(liststruct,1);
	  p=p->next;
	}
	p->item=ScanDepKnot(s);
	p->next=NULL;
	(*n)+=1;
      }
    } while (result!=0);
    fclose(f);
  }
  return list;
}

void usage() {
  printf("depgraph, Creating dependency graph for source code segment,\n");
  printf("(c) Nils Schmeisser, FVTK, FZ-Rossendorf, 1996\n");
  printf(" Usage: 'depgraph' inputfile outputfile\n");
  printf("Semantik: Codesegment: input dependencies -> output dependencies\n");
  printf("  Syntax:      line ::= knotname ': ' identlist ' -> ' identlist\n");
  printf("          identlist ::= ident [ ' ' identlist ]\n");
}

int main(int argc,char **argv) {
  liststruct *list;
  int **imatrix;
  FILE *fout;
  int i,j,n;

  if (argc==3) {
    list=ScanFile(argv[1],&n);
    if (list!=NULL) {
      imatrix=CreateInzMatrix(list,n);
      fout=fopen(argv[2],"w");
//      for (i=0;i<n;i++)
//	fprintf(fout,"%1s",lptr(list,i)->item->name);
//      fprintf(fout,"\n");
      for (i=0;i<n;i++) {
	fprintf(fout,"%4s ",lptr(list,i)->item->name);
	for (j=0;j<n;j++)
	  if (imatrix[i][j]==1)
	    fprintf(fout,"*");
	  else
	    fprintf(fout,".");
	fprintf(fout,"\n");
      }
      fclose(fout);
    } else {
      printf("File not found or file with improper format.\n\n");
      usage();
    }
  } else {
    usage();
  }
  return 0;
}

