/***********************************************************************************************

      File: scheduler.c
     Title: APT Automatic Parallelizing Tool
    Author: Nils Schmeisser
   Version: 1.0
  Rev.Date: 20.2.1996

   Comment: 

***********************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <dep.h>
#include <profiler.h>
#include <cppmacro.h>
#include <apt.h>
#include <aptparser.h>

codeinfostruct *GenerateSync(char *modname,int p1,int p2) {
  codeinfostruct *ci;

  ci=new(codeinfostruct,1);
  ci->type=CODE_STATEMENT;
  ci->name=NULL;
  ci->modname=new(char,strlen(modname)+1);
  strcpy(ci->modname,modname);
  ci->code=new(char,128);
  sprintf(ci->code,"sync(%d,%d); /* synchronizing P%d and P%d */\n",p1,p2,p1,p2);
  ci->node=-1;
  ci->root=0;
  ci->nsucc=0;
  ci->succ=new(void*,MAXDEPNODES);
  ci->npred=0;
  ci->pred=new(void*,MAXDEPNODES);
  ci->line1=-1;
  ci->line2=-1;
  ci->time=0;
  ci->ninput=0;
  ci->input=NULL;
  ci->noutput=0;
  ci->output=NULL;

  return ci;
}

void BuildCode(ciliststruct *cl,int nproc) {
  ciliststruct *l;
  FILE **f;
  int i,j;
  char s[1024];

  printf("#BEGIN_PARALLEL\n");
  for (i=0;i<nproc;i++) {
    l=cl;
    while (l!=NULL) {
      if (l->item->node==i) {
	printf("%s\t/* %s */\n",l->item->code,l->item->name);
	for (j=0;j<l->item->nsyncs;j++) {
	  if (l->item->node!=l->item->syncs[j]) {
	    printf("  SYNC(%d,%d)\n",l->item->node,l->item->syncs[j]);
	  }
	}
      }
      l=l->next;
    }
    if (i<nproc-1)
      printf("#NEXT_TASK\n");
    else
      printf("#END_PARALLEL\n");
  }
}

void MarkNode(codeinfostruct *ci,int node) {
  unsigned int i,j,k;
  int scheduled;
  int mnode;
  codeinfostruct *pred,*p1,*p2;

  if (ci->node<0) {
      printf("%s: eliminating indirect predcessors ",ci->name);
      for (i=0;i<ci->npred;i++) {
	printf("%s ",((codeinfostruct*)(ci->pred[i]))->name);
      }
      printf("\n");
      /* let B,C be preds of A and C pred of B than eliminate C as pred of A */
      for (i=0;i<ci->npred;i++) {
	p1=(codeinfostruct*)(ci->pred[i]);
	if (ci->pred[i]!=NULL) {
	  for (j=0;j<ci->npred;j++) if (i!=j) {
	    p2=(codeinfostruct*)(ci->pred[j]);
	    if (Pred(p1,p2)) {
	      printf("eliminated %s\n",p2->name);
	      for (k=0;k<p2->nsucc;k++) {
		if (p2->succ[k]==p1) {
		  p2->succ[k]=p2->succ[p2->nsucc-1];
		  p2->nsucc--;
		}
	      }
	      ci->pred[j]=ci->pred[ci->npred-1];
	      ci->npred--;
	    }
	  }
	}
      }

    if (ci->npred>1) {

      for (i=0;i<ci->npred;i++) {
	if (((codeinfostruct*)(ci->pred[i]))->node<0)
	  MarkNode((codeinfostruct*)(ci->pred[i]),-1);
      }

      /* now synchronizing pred[]->node */
      mnode=-1;
      for (i=0;i<ci->npred;i++) {
	if (mnode==-1)
	  mnode=((codeinfostruct*)(ci->pred[i]))->node;
	else
	  if (((codeinfostruct*)(ci->pred[i]))->node<mnode)
	    mnode=((codeinfostruct*)(ci->pred[i]))->node;
      }

      printf("%s sync <- ",ci->name);
      ci->node=mnode;
      for (i=0;i<ci->npred;i++) {
	pred=(codeinfostruct*)(ci->pred[i]);
	printf("%s ",pred->name);
	for (j=0;j<ci->npred;j++) if (i!=j) {
	  pred->syncs[pred->nsyncs]=((codeinfostruct*)(ci->pred[j]))->node;
	  pred->nsyncs++;
	}
      }
      printf("\n");
    } else {
      if (node>=0) {
	ci->node=node;

        /*--- einfaches node-splitting, nicht ganz proper, da proc ev. noch belegt ist ---*/

	for (i=0;i<ci->nsucc;i++) {
	  MarkNode((codeinfostruct*)(ci->succ[i]),node); /* % NPROC , zyklische Verteilung */
	}

      } else {
	/* node < 0 signalisiert bottom up */
	if (ci->npred<=0) {
	  printf("could not resolve processor for %s\n",ci->name);
	} else {
	  if (((codeinfostruct*)(ci->pred[0]))->node<0)
	    MarkNode((codeinfostruct*)(ci->pred[0]),-1);
	  ci->node=((codeinfostruct*)(ci->pred[0]))->node;
	}
      }
    }
  } else {
    if (node>=0) {

      /*--- einfaches node-splitting nicht ganz proper, da proc ev. noch belegt ist ---*/

      for (i=0;i<ci->nsucc;i++) {
	MarkNode((codeinfostruct*)(ci->succ[i]),node); /* % NPROC , zyklische Verteilung */
      }
    }
  }
}

void BuildSchedule(ciliststruct *cl,unsigned int nproc) {
  unsigned int nroots,i,j;
  codeinfostruct **roots;
  unsigned int nnode;
  ciliststruct *l;
  char s[1024];
  int needsync;

  APTMessage("building schedule");

  nnode=0;
  roots=new(codeinfostruct*,MAXDEPNODES);

//  l=cl;
//  while (l!=NULL) {
//    printf("%s: >%s<\n",l->item->name,l->item->code);
//    for (i=0;i<l->item->nsucc;i++)
//      if (l->item->succ[i]!=NULL)
//	printf("    -> %s\n",((codeinfostruct*)(l->item->succ[i]))->name);
//    l=l->next;
//  }
    
  APTMessage("searching roots");

  nroots=0;
  l=cl;
  while (l!=NULL) {
    if (l->item->npred==0) {
//      printf("%s is root\n",l->item->name);
      /*--- trivial case nroots <= nproc ---*/
      l->item->node=nnode;
      nnode++;
      roots[nroots]=l->item;
      nroots++;
    } else {
//      for (i=0;i<l->item->npred;i++)
//	printf("%s <- %s\n",l->item->name,((codeinfostruct*)(l->item->pred[i]))->name);
    }
    l=l->next;
  }
    
  sprintf(s,"found %d roots",nroots);
  APTMessage(s);
  
  /* trivial case, there are less roots than processors */

  for (i=0;i<nroots;i++) {
    MarkNode(roots[i],roots[i]->node);
  }

  BuildCode(cl,nproc);

}
